//This file is a part of xboa
//
//xboa is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.
//
//xboa is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with xboa in the doc folder.  If not, see 
//<http://www.gnu.org/licenses/>.
//

#include <Python.h>
#include <structmember.h>
#include <string.h>

#include "Bunchcore.h"


//////////// CONSTRUCTORS AND DESTRUCTORS ////////////
static int  Bunchcore_init_python(Bunchcore* self, PyObject *args, PyObject *kwds)
{
  // int length=-1;
  // if(  PyArg_ParseTuple(args, "i", &length) != 0)
  //   return Bunchcore_init(self, 0);
  // PyErr_Clear();
  return Bunchcore_init(self, 0);
}

static Bunchcore* Bunchcore_shallow_copy(Bunchcore* self)
{
  Bunchcore* copy;
  copy      = malloc(sizeof(Bunchcore));
  if(copy == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return NULL;
  }
  copy->vec = vector_alloc(sizeof(Bunchcore_hit), self->vec->data_length);
  int i=0;
  for(i=0; i<self->vec->data_length; i++)
  {
      Bunchcore_hit* bc_hit_copy = vector_el(copy->vec, i);
      Bunchcore_hit* bc_hit_self = vector_el(self->vec, i);
      bc_hit_copy->hit           = bc_hit_self->hit;
      bc_hit_copy->hitcore       = bc_hit_self->hitcore;
      Py_INCREF(bc_hit_copy->hit);
  }
  return copy;
}

static Bunchcore* bc_realloc(Bunchcore* self, int new_length)
{
  int i, old_length = length(self);
  if(vector_realloc(self->vec, new_length) == NULL) { return NULL;}
  for(i=old_length; i<new_length; i++) 
  { 
    Bunchcore_hit * bc_hit = vector_el(self->vec, i); 
    bc_hit->hit = NULL; 
    bc_hit->hitcore = NULL;
  }
  return self;
}

static int  Bunchcore_init(Bunchcore* self, int len)
{
  //python does the malloc for Bunchcore
  //I need to make sure children of Bunchcore are initialised
  self->vec  = vector_alloc(sizeof(Bunchcore_hit), 0);
  if(self->vec == NULL) return -1;
  return 0;
}

static void Bunchcore_dealloc(Bunchcore * self)
{
  size_t i;
  for(i=0; i<vector_size(self->vec); i++)
  {
    Bunchcore_hit* bc_hit = vector_el(self->vec, i);
    if (bc_hit->hit != NULL) 
      Py_DECREF(bc_hit->hit);
    if (bc_hit->hitcore  != NULL) 
      Py_DECREF(bc_hit->hitcore);
  }
  vector_free(self->vec);
}


///////// SEQUENCE METHODS /////////////
//Sequence methods are the ones that make the Bunchcore look like a python-style list
//So these are all the things like referencing objects in the list, concatenations, etc
//Not sure I need all these methods, might be better to leave some out and do them at 
//Bunch level...
//
//First C methods ignoring python API
//WARNING - bound checking, index slicing etc is done in python API calls
//so segmentation errors abound!

//There's a whole load of stuff commented out at the end - probably works, not checked

//set item - if required, caller must handle Py_INCREF on hit and do bound checking
static void set_item_bc_hit(Bunchcore* self, PyObject* a_hit, Hitcore* a_hitcore, int pos, int* ierr)
{
  if(pos >= length(self) ) 
    if(bc_realloc(self, pos+1) == NULL) {*ierr = -1; return;}
  Bunchcore_hit* new = vector_el_bc(self->vec, pos);
  new->hitcore = a_hitcore;
  if(new->hit != NULL) {
    Py_DECREF(new->hit);
    Py_DECREF(new->hitcore);
  }
  new->hit     = a_hit;
  if (new->hit != NULL)
    Py_INCREF(new->hitcore);
    Py_INCREF(new->hit);
}

static PyObject* length_python(PyObject* self, PyObject* args) 
{ 
  Bunchcore * bc = (Bunchcore*)self;
  int len = 0;
  if(bc->vec->object_size > 0)
    len = (int)(bc->vec->data_length/bc->vec->object_size);
  return Py_BuildValue("i", len);
}

static Py_ssize_t len_python(PyObject* self) //len() function wants Py_ssize_t in return
{
  Py_ssize_t len = 0;
  Bunchcore * bc = (Bunchcore*)self;
  if(bc->vec->object_size > 0) len = bc->vec->data_length/bc->vec->object_size;
  return len;
}

//Now Python wrappers for (some of) above functions
//Set item - call sign is (self, Hit, Hitcore, int)
static PyObject* set_item_python(PyObject* self, PyObject* args)
{
  int        pos;
  PyObject*  a_hit;
  Hitcore*   a_hitcore;
  Bunchcore* me = (Bunchcore*)self;
  if(  PyArg_ParseTuple(args, "OOi", &a_hit, &a_hitcore, &pos) != 0)
  {
    if(pos < 0) pos = pos + length(me);
    set_item_bc_hit(me, a_hit, a_hitcore, pos, &pos); //Py_INCREF handled here
    if(pos == -1)
    {
      PyErr_SetString(PyExc_IndexError,  "Internal error in Bunchcore.set_item - out of memory?");
      return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
  }
  PyErr_SetString(PyExc_KeyError,  "Failed to parse variable in function Bunchcore.set_item");
  return NULL; //raise an exception
}

static PyObject* get_item_python(PyObject* self, PyObject* args)
{
  int        pos;
  Bunchcore* me = (Bunchcore*)self;
  if(  PyArg_ParseTuple(args, "i", &pos) != 0)
  {
    if(pos < 0) pos += length(me);
    Bunchcore_hit* bc_hit = vector_el_bc(me->vec, pos);
    if(bc_hit == NULL)
    {
      PyErr_SetString(PyExc_IndexError,  "Index out of range in Bunchcore.set_item");
      return NULL;
    }
    if(bc_hit->hit == NULL) {Py_INCREF(Py_None); Py_INCREF(Py_None); return Py_BuildValue("(OO)", Py_None, Py_None);}
    return Py_BuildValue("(OO)", bc_hit->hit, bc_hit->hitcore);
  }
  PyErr_SetString(PyExc_KeyError,  "Failed to parse variable in function Bunchcore.set_item");
  return NULL; //raise an exception
}

//////////////// MOMENTS //////////////////////
static double get_moment_char(Bunchcore*  self, const char** axes, double* means, int n_axes, int* ierr)
{
  //turn list of axes into a list of function calls
  hc_get_dbl_function   functions[n_axes];
  hc_get_dbl_function   hc_get_total_weight = hc_get_dbl_function_by_string("weight");
  int i=0, j=0;
  for(i=0; i<n_axes; i++)
  {
    functions[i] = hc_get_dbl_function_by_string(axes[i]);
    if(functions[i] == NULL) {*ierr = -1; return 0.;}
  }

  //extract weights
  double w = 0.;
  for(i=0; i<length(self); i++) {
    double dw = hc_get_total_weight( ((Bunchcore_hit*)vector_el(self->vec, i) )->hitcore );
    w += dw;
  }

  //extract moments
  double m = 0.;
  for(i=0; i<length(self); i++)
  {
    double my_m = hc_get_total_weight( ((Bunchcore_hit*)vector_el(self->vec, i) )->hitcore ) / w;
    for(j=0; j<n_axes; j++) //could save a function look-up for duped axes...
      my_m *= functions[j]( ((Bunchcore_hit*)vector_el(self->vec, i) )->hitcore ) - means[j]; 
    m += my_m;
  }
  return m;
}

static void covariance_matrix_char(Bunchcore*  self, const char* axes[], double* means, int n_axes, double* target, int* ierr)
{
  hc_get_dbl_function hc_get_total_weight = hc_get_dbl_function_by_string("weight");
  int i=0, j=0, k=0;
  hc_get_dbl_function functions[n_axes];
  int n_hits = length(self);
  Hitcore* hits[n_hits];
  for(i=0; i<n_hits; i++)
    hits[i] = ((Bunchcore_hit*)vector_el_bc(self->vec, i) )->hitcore;

  for(i=0; i<n_axes; i++)
  {
    functions[i] = hc_get_dbl_function_by_string(axes[i]);
    if(functions[i] == NULL) {*ierr = -1; return;}
  }

  double w = 0.;
  for(i=0; i<n_hits; i++)
    w += hc_get_total_weight( hits[i] );

  for(j=0; j<n_axes; j++) //could save a function look-up for duped axes...
    for(k=0; k<n_axes; k++)
      target[j+k*n_axes] = 0.;

  for(i=0; i<n_hits; i++)
  {
    for(j=0; j<n_axes; j++) //could save a function look-up for duped axes...
    {
      double my_m = hc_get_total_weight( hits[i] )*(functions[j]( hits[i] )-means[j])/w; 
      for(k=0; k<n_axes; k++)
        target[j+k*n_axes] += my_m*(functions[k]( hits[i] ) - means[j]);
    }
  }
}

static void make_array_from_char_list(PyObject* py_list, const char*** string_list, int* list_size)
{
  //gotta love C - level 3 pointer! (must have Fireball spell by now?)
  if(!py_list || !PyList_Check(py_list)) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to parse argument list of strings in Bunchcore");
    *string_list = NULL;
    return;
  }

  int i;
  *list_size   = PyList_Size(py_list);
  *string_list = malloc(sizeof(const char*[*list_size]));
  if(*string_list == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return;
  }
  for(i=0; i<*list_size; i++) 
  {
    PyObject* my_string = PyList_GetItem(py_list, i);
    if( PyString_Check(my_string) ) (*string_list)[i] = PyString_AsString(my_string);
    else
    {
        PyErr_SetString(PyExc_KeyError,  "Failed to parse argument list of strings in Bunchcore");
        free(*string_list);
        *string_list = NULL;
        return;
    }
  }
}

static void make_array_from_float_list(PyObject* py_list, double** double_list, int* list_size)
{
  //gotta love C - level 3 pointer! (must have Fireball spell by now?)
  if(!py_list || !PyList_Check(py_list))
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to parse argument list of strings in Bunchcore");
    *double_list = NULL;
    return;
  }

  int i;
  *list_size   = PyList_Size(py_list);
  *double_list = malloc(sizeof(double[*list_size]));
  if(*double_list == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return;
  }

  for(i=0; i<*list_size; i++) 
  {
    PyObject* my_float =  PyList_GetItem(py_list, i);
    if(PyFloat_Check(my_float)) (*double_list)[i] = PyFloat_AsDouble(my_float) ;
    else
    {
        PyErr_SetString(PyExc_KeyError,  "Failed to parse argument list of floats in Bunchcore");
        free(*double_list);
        *double_list = NULL;
        return;
    }
  }
}


static PyObject* make_lists_from_double_array(double values[], int list_size)
{
  int i=0,j=0,k=0,err=0;
  PyObject* the_list  = PyList_New(list_size);
  for(i=0; i<list_size; i++) 
  {
    PyObject* a_list  = PyList_New(list_size);
    err = PyList_SetItem(the_list, i, a_list);
    for(j=0; j<list_size; j++)
    {
      PyObject* a_float = Py_BuildValue("d", values[k]);
      err = PyList_SetItem(a_list, j, a_float);
      k++;
    }
  }
  if (err == 0) {} // kill gcc warning
  return the_list;
}

static PyObject* get_moment_python(PyObject*  self, PyObject* args)
{
  PyObject* py_list_axes, *py_list_means;
  int       list_size1, list_size2, i=0;
  const char***  string_list = malloc(sizeof(const char**));
  double**       double_list = malloc(sizeof(double*));
  if(string_list == NULL || double_list == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return NULL;
  }

  if( PyArg_ParseTuple(args, "OO", &py_list_axes, &py_list_means) == 0) return NULL;
  make_array_from_char_list (py_list_axes,  string_list, &list_size1); //allocates string_list
  if(*string_list == NULL) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_moment");
    free(string_list);
    free(double_list);
    return NULL;
  }
  make_array_from_float_list(py_list_means, double_list, &list_size2);
  if(*double_list == NULL) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_moment");
    free(*string_list);
    free(string_list);
    free(double_list);
    return NULL;
  }
  if(list_size1 != list_size2) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_moment");
    free(*string_list);
    free(*double_list);
    free(string_list);
    free(double_list);
    return NULL;
  }
  double moment = get_moment_char( (Bunchcore*)self, *string_list, *double_list, list_size1, &i);
  free(*double_list); free(*string_list); free(string_list); free(double_list);
  if(i < 0) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_moment");
    return NULL;
  }
  PyErr_Clear();
  return Py_BuildValue("d", moment);
}

static PyObject* covariance_matrix_python(PyObject*  self, PyObject* args)
{
  PyObject* py_list_axes, *py_list_means;
  int       list_size1, list_size2, i=0;
  const char***  string_list = malloc(sizeof(const char**));
  double**       double_list = malloc(sizeof(double*));
  if(string_list == NULL || double_list == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return NULL;
  }

  if( PyArg_ParseTuple(args, "OO", &py_list_axes, &py_list_means) == 0) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_covariance_matrix");
    return NULL;
  }
  make_array_from_char_list (py_list_axes,  string_list, &list_size1); //allocates string_list
  if(*string_list == NULL) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_covariance_matrix");
    free(string_list);
    free(double_list);
    return NULL;
  }

  make_array_from_float_list(py_list_means, double_list, &list_size2); //allocates double_list or returns NULL
  if(*double_list == NULL) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_covariance_matrix");
    free(*string_list);
    free(string_list);
    free(double_list);
    return NULL;
  }
  if(list_size1 != list_size2) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_covariance_matrix");
    free(*string_list);
    free(*double_list);
    free(string_list);
    free(double_list);
    return NULL;
  }

  double* target = malloc(sizeof(double[list_size1*list_size1]));
  if(target == NULL) 
  {
    PyErr_SetString(PyExc_KeyError,  "Failed to allocate memory in Bunchcore");
    return NULL;
  }
  covariance_matrix_char( (Bunchcore*)self, *string_list, *double_list, list_size1, target, &i);
  PyObject* cov_matrix = make_lists_from_double_array(target, list_size1);

  free(*double_list); free(*string_list); free(string_list); free(double_list);
  if(i < 0) {
    PyErr_SetString(PyExc_KeyError,  "Failed while calculating moment in Bunchcore::get_covariance_matrix");
    return NULL;
  }
  free(target);
  PyErr_Clear();

  return cov_matrix;
}

static PyObject* _cut_double_python(PyObject* self, PyObject* args) {
  char* get_variable = NULL;
  PyObject* comp = NULL;
  double cut_value = 0.;
  char* weight_variable = NULL;
  if (PyArg_ParseTuple(args, "sOds", &get_variable, &comp, &cut_value, &weight_variable) == 0)
      return NULL;
  int is_local = strcmp("local_weight", weight_variable);
  if (is_local == 0)
      if (strcmp("global_weight", weight_variable) == 0) {
          PyErr_SetString(PyExc_ValueError, "Attempting to cut with odd set value");
          return NULL;
      }
  if (_cut_double((Bunchcore*)self, get_variable, comp, cut_value, is_local) == 0) {
      PyErr_SetString(PyExc_ValueError, "Failed to apply cut");
      return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

static int _cut_double(Bunchcore* self, const char* get_variable,
                  PyObject* comp, const double cut_value,
                  const int is_local) {
  hc_get_dbl_function get_var = hc_get_dbl_function_by_string(get_variable);
  if (get_var == NULL) {
      return 0;
  }
  int my_length = length(self);
  int i = 0;
  for (i = 0; i < my_length; ++i) {
    Hitcore* hit = ((Bunchcore_hit*)vector_el(self->vec, i) )->hitcore;
    double variable = get_var(hit);
    PyObject* comp_out = PyObject_CallFunction(comp, "dd", variable, cut_value);
    if (comp_out == NULL) {
        return 0;
    }
    int comp = 0;
    int ok = PyArg_Parse(comp_out, "i", &comp);
    if (!ok) {
        return 0;
    }
    if (comp == 1) {
        if (is_local == 0)    
            hc_set_local_weight(hit, 0.);
        else
            hc_set_global_weight(hit, 0.);
    }
  }
  return 1;
}

//      hit_value_list = self.list_get_hit_variable([variable])[0]
//      for i in range ( len(hit_value_list) ):
//        if comparator(hit_value_list[i], cut_value): 
//          self.__hits[i].set(set_var, 0)

//////////////// PYTHON TYPE DEFINITION ///////////////
static PyMemberDef Bunchcore_members[] = {
{NULL},
};

static PyMethodDef Bunchcore_methods[] = {
{"set_item",  (PyCFunction)set_item_python,   METH_VARARGS, "Set i^th item"},
{"get_item",  (PyCFunction)get_item_python,   METH_VARARGS, "Get i^th item"},
{"length",    (PyCFunction)length_python,     METH_VARARGS, "Report the length of the Bunchcore"},
{"moment",    (PyCFunction)get_moment_python, METH_VARARGS, "Get moment from a list of strings"},
{"covariance_matrix", (PyCFunction)covariance_matrix_python, METH_VARARGS, "Get covariance matrix from a list of strings"},
{"_cut_double", (PyCFunction)_cut_double_python, METH_VARARGS, "Apply cut inner loop. Returns NULL on error or if get_variable is an int"},
{NULL}
};

static PySequenceMethods Bunchcore_as_seq = {
  (lenfunc)len_python, //lenfunc PySequenceMethods.sq_length
  0, //binaryfunc PySequenceMethods.sq_concat
  0, //ssizeargfunc PySequenceMethods.sq_repeat
  0, //ssizeargfunc PySequenceMethods.sq_item
  0, //ssizeobjargproc PySequenceMethods.sq_ass_item
  0, //objobjproc PySequenceMethods.sq_contains
  0, //binaryfunc PySequenceMethods.sq_inplace_concat
  0, //ssizeargfunc PySequenceMethods.sq_inplace_repeat
};

static PyTypeObject BunchcoreType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "Bunchcore.Bunchcore",     /*tp_name*/
    sizeof(Bunchcore),         /*tp_basicsize*/
    sizeof(Bunchcore_hit),  /*tp_itemsize*/
    (destructor)Bunchcore_dealloc,     /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    &Bunchcore_as_seq,          /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
    "Bunchcore objects",           /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    Bunchcore_methods,           /* tp_methods */
    Bunchcore_members,           /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)Bunchcore_init_python,      /* tp_init */
    0,                         /* tp_alloc */
    0,                 /* tp_new */
};

PyMODINIT_FUNC initBunchcore(void)
{
  import_hitcore();

  BunchcoreType.tp_new = PyType_GenericNew;
  if (PyType_Ready(&BunchcoreType) < 0) return;

  PyObject* m = Py_InitModule3("Bunchcore", NULL, "Core module wrapped by Bunch");
  if (m == NULL) return;

  Py_INCREF(&BunchcoreType);
  PyModule_AddObject(m, "Bunchcore", (PyObject *)&BunchcoreType);
}

void import_hitcore(void) {
  PyObject* hc_module = PyImport_ImportModule("xboa.core.Hitcore");
  if(hc_module == NULL) {return;}
  else
  {
    PyObject *module_dict  = PyModule_GetDict(hc_module);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "HITCORE_C_API");
    if (PyCObject_Check(c_api_object)) 
    {
      hitcore_api = (void **)PyCObject_AsVoidPtr(c_api_object);
      hc_get_dbl_function_by_string = hitcore_api[0];
      hc_set_local_weight           = hitcore_api[1]; 
      hc_set_global_weight          = hitcore_api[2]; 
      vector_alloc                  = hitcore_api[3];
      vector_realloc                = hitcore_api[4];
      vector_init                   = hitcore_api[5];
      vector_insert                 = hitcore_api[6];
      vector_el                     = hitcore_api[7];
      vector_el_bc                  = hitcore_api[8];
      vector_size                   = hitcore_api[9];
      vector_free                   = hitcore_api[10];
    }
  }
}

//Dummy function to remove compiler warnings
//sorry
void no_warn(void)
{
  return;
  Bunchcore_shallow_copy(NULL);
}

/*
static Bunchcore* concat(Bunchcore* a, Bunchcore* b) 
{
  Bunchcore* bc = malloc(sizeof(Bunchcore));
  Bunchcore_init(bc, a->length+b->length);
  int i;
  for(i=0; i<a->length; i++)
    set_item_bc_hit(bc, a->first[i].hit, a->first[i].hitcore, i);
  for(i=0; i<b->length; i++)
    set_item_bc_hit(bc, b->first[i].hit, b->first[i].hitcore, i+a->length);
  return bc;
}

static void inplace_concat(Bunchcore* a, Bunchcore* b)  //pretty sure this does not work!
{
  Bunchcore* copy = Bunchcore_shallow_copy(a);
  Bunchcore_init(a, a->length+b->length);
  int i;
  for(i=0; i<copy->length; i++)
    set_item_bc_hit(a, a->first[i].hit, a->first[i].hitcore, i);
  for(i=0; i<copy->length; i++)
    set_item_bc_hit(a, b->first[i].hit, b->first[i].hitcore, i+a->length);
  Bunchcore_dealloc(copy);
}

//Return a Bunchcore with nrepeats copies of a
static Bunchcore* repeat(Bunchcore* a, int nrepeats)
{
  Bunchcore* bc = malloc(sizeof(Bunchcore));
  Bunchcore_init(bc, a->length*nrepeats);
  int i,j,k=0;
  for(i=0; i<a->length; i++)
    for(j=0; j<nrepeats; j++)
    {
      set_item_bc_hit(a, a->first[k].hit, a->first[k].hitcore, i);
      k++;
    }
  return bc;
}

//Rebuild a as a Bunch with nrepeats copies of a before function call
static void inplace_repeat(Bunchcore* a, int nrepeats)  //pretty sure this does not work!
{
  Bunchcore* copy = Bunchcore_shallow_copy(a);
  Bunchcore_init(a, a->length*nrepeats);
  int i,j,k=0;
  for(i=0; i<copy->length; i++)
    for(j=0; j<nrepeats; j++)
    {
      set_item_bc_hit(a, a->first[k].hit, a->first[k].hitcore, i);
      k++;
    }
}

//Return the hit at index *blah* (no bound checking)
static PyObject* get_hit(Bunchcore* self, int index)
{return ( (Bunchcore_hit*)self->first+index*BunchcoreType.tp_itemsize )->hit;}

//Return the hitcore associated with the hit at index
static Hitcore* get_hitcore(Bunchcore* self, int index)
{return ( (Bunchcore_hit*)self->first+index*BunchcoreType.tp_itemsize )->hitcore;}

//Return 0 if self contains hit, else return -1
static int contains_hit(Bunchcore* self, PyObject* hit)
{
  int i;
  for(i=0; i<self->length; i++)
    if( ((Bunchcore_hit*)self->first+i*BunchcoreType.tp_itemsize )->hit == hit) return 0;
  return -1;
}
*/

